import system.io
import .imp.syntax
import .imp.common

-- class gen (a : Type) :=
--   (fn : int → a)

-- meta def generate (a : Type) [gen a] : io a :=
--   do i ← random_int,
namespace imp

variable [io.interface]

meta def random_int (bound : option int := none) : io int :=
do res ← io.cmd { cmd := "bash", args := ["-c", "'echo $RANDOM'"]},
   let res := list.take (res.length - 1) res.to_list,
   match from_string res with
   | none := io.fail "unable to generate a random number"
   | some v :=
    match bound with
    | none := return v
    | some bound := return (v % bound)
   end
  end

-- instance char.gen : gen char :=
--   ⟨ λ i, char.of_nat (32 + (abs i).to_nat) ⟩

meta def rand_char : io char :=
  do i ← random_int (some $ 127 - 32),
     return $ char.of_nat (32 + (abs i).to_nat)

meta def range (start : int) : list int :=
  list.repeat start start.to_nat

meta def rand_list {α : Type} (gen : int → io α) (max : int) : io (list α) :=
  do i ← random_int max,
     monad.mapm gen (range max)

meta def rand_var : io string :=
  do i ← random_int (some 10),
     return $ "x" ++ to_string i

meta def rand_bool : io bool :=
  do i ← random_int,
     if i % 0 = 2
     then return tt
     else return ff

meta def rand_val : io val :=
  do i ← random_int (some 3),
    if i = 0
    then val.int <$> random_int (some 20)
    else if i = 1
    then val.bool <$> rand_bool
    else if i = 2
    then do cs ← rand_list (fun _, rand_char) 50,
            return $ val.str $ cs.as_string
    else io.fail "rand_val"

meta def rand_op1 : io unary_op :=
  do i ← random_int (some 2),
     if i = 0
     then pure unary_op.neg
     else if i = 1
     then pure unary_op.not
     else io.fail "rand_op1"

meta def rand_binary_op : io binary_op :=
do i ← random_int (some 10),
  match i with
  | 0 := pure $ binary_op.add
  | 1 := pure $ binary_op.sub
  | 2 := pure $ binary_op.mul
  | 3 := pure $ binary_op.div
  | 4 := pure $ binary_op.mod
  | 5 := pure $ binary_op.eq
  | 6 := pure $ binary_op.lt
  | 7 := pure $ binary_op.le
  | 8 := pure $ binary_op.conj
  | 9 := pure $ binary_op.disj
  | _ := io.fail "rand_binary_op"
  end

meta def rand_expr : int → io expr
| size :=
  if size <= 1
  then do i ← random_int (some 2),
       match i with
       | 0 := expr.val <$> rand_val
       | 1 := expr.var <$> rand_var
       | _ := io.fail "rand_expr in size <= 1"
       end
  else do i ← random_int (some 6),
       match i with
       | 0 := expr.val <$> rand_val
       | 1 := expr.var <$> rand_var
       | 2 := expr.unary_op <$> rand_op1 <*> rand_expr (size - 1)
       | 3 := expr.binary_op <$>
                rand_binary_op <*>
                rand_expr (size / 2) <*>
                rand_expr (size / 2)
       | 4 := expr.length <$> rand_expr (size - 1)
       | 5 := expr.index <$> rand_expr (size / 2) <*> rand_expr (size / 2)
       | _ := io.fail "rand_expr"
       end

meta def rand_stmt : int → io stmt
| size :=
  if size <= 1
  then do i ← random_int (some 4),
       match i with
       | 0 := pure $ stmt.nop
       | 1 := stmt.set <$> rand_var <*> rand_expr 7
       | 2 := stmt.alloc <$> rand_var <*> rand_expr 5 <*> rand_expr 5
       | 3 := stmt.write <$> rand_var <*> rand_expr 5 <*> rand_expr 5
       | _ := io.fail "rand_stmt in size <= 1"
       end
  else do i ← random_int (some 7),
       match i with
       | 0 := pure stmt.nop
       | 1 := stmt.set <$> rand_var <*> rand_expr 7
       | 2 := stmt.alloc <$> rand_var <*> rand_expr 5 <*> rand_expr 5
       | 3 := stmt.write <$> rand_var <*> rand_expr 5 <*> rand_expr 5
       | 4 := stmt.seq <$> rand_stmt (size / 2) <*> rand_stmt (size / 2)
       | 5 := stmt.ifelse <$> rand_expr 5 <*> rand_stmt (size / 2) <*> rand_stmt (size /2)
       | 6 := stmt.while <$> rand_expr 5 <*> rand_stmt (size - 1)
       | _ := io.fail "rand_stmt"
       end

meta def rand_func (size : int) : io func :=
  func.mk <$> rand_var <*>
              rand_list (fun _, rand_var) size <*>
              rand_stmt size <*>
              rand_expr size

meta def rand_prog (size : int) : io prog :=
  prog.mk <$>
    rand_list rand_func size <*>
    rand_stmt size <*>
    rand_expr size

meta def gen_int : io int :=
  do b ← rand_bool,
     i ← random_int (some 100),
     pure $ (if b then 1 else -1) * i

meta def rand_elt {a:Type} [inhabited a] (xs : list a) : io a :=
  do i ← random_int (some $ xs.length),
     pure $ (xs.nth i.to_nat).get_or_else (xs.head)

-- (* NOTE : typ only heuristic to help avoid generating bogus code *)

inductive typ
| int
| bool
| addr

meta def has_typ : typ → val → bool
| typ.int (val.int _) := tt
| typ.bool (val.bool _) := tt
| typ.addr (val.addr _) := tt
| _ _ := ff

meta def rand_typ : io typ :=
  do i ← random_int (some 3),
  match i with
  | 0 := pure typ.int
  | 1 := pure typ.bool
  | 2 := pure typ.addr
  | _ := io.fail "rand_typ"
  end

meta def gen_op1 : typ → typ → io unary_op
| typ.int typ.int := pure unary_op.neg
| typ.bool typ.bool := pure unary_op.not
| _ _ := io.fail "gen_op1"

meta def gen_op2 : typ → typ → typ → io binary_op
| typ.int typ.int typ.int :=
      do i ← random_int (some 5),
      match i with
      | 0 := pure binary_op.add
      | 1 := pure binary_op.sub
      | 2 := pure binary_op.mul
      | 3 := pure binary_op.div
      | 4 := pure binary_op.mod
      | _ := io.fail "gen_op2 i i i"
      end
| typ.int typ.int typ.bool :=
     do i ← random_int (some 3),
      match i with
      | 0 := pure binary_op.eq
      | 1 := pure binary_op.conj
      | 2 := pure binary_op.disj
      | _ := io.fail "gen_op2 i i b"
      end
| typ.bool typ.bool typ.bool :=
      do i ← random_int (some 3),
      match i with
      | 0 := pure binary_op.eq
      | 1 := pure binary_op.conj
      | 2 := pure binary_op.disj
      | _ := io.fail "gen_op2 b b b"
      end
| _ _ typ.bool := pure binary_op.eq
| _ _ _ := io.fail "gen_op2"

instance : inhabited (string × val) :=
inhabited.mk ("", val.int 0)

meta def select_rand_var (t : typ) (s : store) : io string :=
  let tvs := s.filter (fun p, has_typ t (prod.snd p))
  in prod.fst <$> rand_elt tvs

meta def gen_expr : typ → store → heap → int → io imp.expr
| typ.int s h n :=
      if n <= 1 then
        do i ← random_int (some 2),
        match i with
        | 0 := do i ← random_int, pure $ expr.val (val.int i)
        | 1 := expr.var <$> select_rand_var typ.int s
        | _ := io.fail "gen_expr i 1"
        end
      else
        do i ← random_int (some 6),
        match i with
        | 0 := expr.val <$> val.int <$> gen_int
        | 1 := expr.var <$> select_rand_var typ.int s
        | 2 := expr.unary_op <$>
                    gen_op1 typ.int typ.int
                    <*> gen_expr typ.int s h (n - 1)
        | 3 := expr.binary_op <$> gen_op2 typ.int typ.int typ.int
                    <*> gen_expr typ.int s h (n / 2)
                    <*> gen_expr typ.int s h (n / 2)
        | 4 := expr.length <$> (gen_expr typ.addr s h (n - 1))
        | 5 := /- unsound -/
               expr.index <$> gen_expr typ.addr s h (n / 2) <*>
                              gen_expr typ.int s h (n / 2)
        | _ := io.fail "gen_expr i n"
        end
| typ.bool s h n :=
  if n <= 1 then
    do i ← random_int (some 2),
    match i with
    | 0 := expr.val <$> (val.bool <$> rand_bool)
    | 1 := expr.var <$> select_rand_var typ.bool s
    | _ := io.fail "gen_expr b 1"
    end
  else
    do i ← random_int (some 5),
    match i with
    | 0 := expr.val <$> (val.bool <$> rand_bool)
    | 1 := expr.var <$> select_rand_var typ.bool s
    | 2 := expr.unary_op <$>
            gen_op1 typ.bool typ.bool
            <*> gen_expr typ.bool s h (n - 1)
    | 3 :=
      do i ← random_int (some 3),
      match i with
      | 0 := expr.binary_op <$> gen_op2 typ.int typ.int typ.bool
                  <*> gen_expr typ.int s h (n / 2)
                  <*> gen_expr typ.int s h (n / 2)
      | 1 := expr.binary_op  <$> gen_op2 typ.bool typ.bool typ.bool
                        <*> gen_expr typ.int s h (n / 2)
                        <*> gen_expr typ.int s h (n / 2)
      | 2 := expr.binary_op binary_op.eq <$>
                        (rand_typ >>= (fun t, gen_expr t s h (n / 2)))
                        <*> (rand_typ >>= fun t, gen_expr t s h (n / 2))

      | _ := io.fail "gen_expr b n 3"
      end
    | 4 := /- unsound -/
               expr.index <$> gen_expr typ.addr s h (n / 2) <*> gen_expr typ.int s h (n / 2)
    | _ := io.fail "gen_expr b n"
   end
| typ.addr s h n :=
      if n <= 1 then
        expr.var <$> select_rand_var typ.addr s
      else
        do i ← random_int (some 2),
        match i with
        | 0 := expr.var <$> select_rand_var typ.addr s
        | 1 := /- unsound -/
               expr.index <$> gen_expr typ.addr s h (n / 2)
                          <*> gen_expr typ.int s h (n / 2)
        | _ := io.fail "gen_expr a n"
        end

-- meta def char_of_digit d =
--   Char.chr (Char.code '0' + d)

-- meta def rand_var () =
--   [ 'x'
--   ; char_of_digit (Random.int 10)
--   ]

meta def gen_stmt : store → heap → int → io stmt
| s h n :=
if n <= 1 then
    do i ← random_int (some 3),
    match i with
    | 0 := stmt.set <$> rand_var
                <*> (rand_typ >>= fun t, gen_expr t s h 7)
    | 1 := stmt.alloc <$> rand_var
                      <*> gen_expr typ.int s h 5
                      <*> (rand_typ >>= fun t, gen_expr t s h 5)
    | 2 := stmt.write <$> select_rand_var typ.addr s
                    <*> gen_expr typ.int s h 5
                    <*> (rand_typ >>= fun t, gen_expr t s h 5)
    | _ := io.fail "gen_stmt 1"
    end
  else
    do i ← random_int (some 3),
    match i with
    | 0 := stmt.set <$> rand_var
                    <*> (rand_typ >>= fun t, gen_expr t s h 7)
    | 1 := stmt.alloc <$> rand_var
                  <*> gen_expr typ.int s h 5
                  <*> (rand_typ >>= fun t, gen_expr t s h 5)
    | 2 := stmt.write <$> select_rand_var typ.addr s
                      <*> gen_expr typ.int s h 5
                      <*> (rand_typ >>= fun t, gen_expr t s h 5)
    | 3 := stmt.seq <$> gen_stmt s h (n / 2)
                    <*> gen_stmt s h (n / 2)
    | 4 := stmt.ifelse <$> gen_expr typ.bool s h 5
                   <*> gen_stmt s h (n / 2)
                   <*> gen_stmt s h (n / 2)
    | 5 := stmt.while <$> gen_expr typ.bool s h 5
                      <*> gen_stmt s h (n - 1)
    | _ := io.fail "gen_stmt 1"
    end


open stmt
open expr

meta def gen (n : int) : io stmt :=
  stmt.seq ( stmt.set "x0" $ expr.val (val.int 0)) <$>
  stmt.seq ( stmt.set "x1" $ expr.val (val.bool false)) <$>
  stmt.seq ( alloc "x2" (expr.val (val.int 3)) (expr.val (val.int 0))) <$>
  (gen_stmt [ ("x0", val.int 0),
              ("x1", val.bool ff),
             ("x2", val.addr 0)]
            [val.int 3, val.int 0, val.int 0, val.int 0]
            n)

end imp
