import .syntax
import .decidable_eq
import ..unsafe_io
import .has_to_string

def string.index : string → nat → option char :=
fun s n, s.to_list.nth n

def from_char (c : char) : option int :=
if '0' <= c ∧ c <= '9'
then some (c.to_nat - 48)
else none

def from_string (cs : list char): option int :=
  let (mul, cs) := (match cs with
  | ('-'::cs) := (-1, cs)
  | _ := (1, cs)
  end : (int × list char))in
  do digits ← monad.mapm from_char cs,
     return $ mul * digits.foldl (fun s i, s * 10 + i) 0

namespace imp

-- Stores
def update (x : string) (v : val) : store → store
| [] := [(x, v)]
| ((x', v') :: rest) :=
  if x = x'
  then (x, v) :: rest
  else (x', v') :: (update rest)

def updates : list string → list val →  store → option store
| [] [] s := some s
| (x :: xs') (v :: vs') s := updates xs' vs' (update x v s)
| _ _ _ := none

def store.lookup (x : string) : store → option val
| [] := none
| ((x', v) :: rest) :=
  if x = x'
  then some v
  else store.lookup rest

def is_nop (p : stmt) : decidable (p = stmt.nop) :=
begin
  refine (if H : p = stmt.nop then _ else _),
  apply decidable.is_true,
  assumption,
  apply decidable.is_false,
  assumption,
end

def heap.empty : heap := []

def copy : nat → val → list val
| 0 v := []
| (n + 1) v := v :: copy n v

def heap.alloc (h : heap) (i : ℤ) (v : val) : heap :=
h ++ (val.int i :: copy (int.to_nat i) v)

def heap.read : heap → int → option val
| [] _ := none
| (v :: vs) 0 := some v
| (v :: vs) i := heap.read vs (i - 1)

def heap.write : heap → ℤ → val → option heap
| []  _ _ := none
| (x :: xs) 0 v := some (v :: xs)
| (x :: xs) i v :=
  match heap.write xs (i - 1) v with
  | some h' := some (x :: h')
  | none := none
  end

-- Common
def heap.size : heap → ℤ
| [] := 0
| (_ :: rest) := 1 + heap.size rest.

open stmt

mutual def stmt.size, stmt_size_args
with stmt.size : stmt → nat
| nop := 0
| (set _ _) := 0
| (alloc _ _ _) := 0
| (write _ _ _) := 0
| (ifelse _ p1 p2) := nat.succ (stmt.size p1 + stmt.size p2)
| (while _ p1) := nat.succ (stmt.size p1)
| (seq p1 p2) := nat.succ (stmt.size p1 + stmt.size p2)
| (call res f args) := nat.succ (stmt_size_args args)
with stmt_size_args : list expr → nat
| [] := 0
| (arg :: args) := 1 + stmt_size_args args

inductive result : Type
| done : heap → val → result
| timeout : store → heap → stmt → expr → result
| stuck : store → heap → stmt → expr → result


/- Environments -/

@[reducible] def env : Type := list func

def locate : env → string → option func
| [] _ := none
| (f :: rest) x :=
match f with
| (func.mk x' _ _ _) :=
  if x = x'
  then some f
  else locate rest x
end.

/- External Calls -/
inductive extcall_spec :
  string → list val → heap → val → heap → Prop
| print_spec :
    forall v h v',
      extcall_spec
        "print_val" (v :: []) h
        v' h
| flush_spec :
    forall h v',
      extcall_spec
        "flush" [] h
        v' h
| sleep_spec :
    forall i h v',
      extcall_spec
        "sleep" (val.int i :: []) h
        v' h
| read_bool_spec :
    forall h b,
      extcall_spec
        "read_bool" [] h
        (val.bool b) h
| read_int_spec :
    forall h i,
      extcall_spec
        "read_int" [] h
        (val.int i) h
| read_str_spec :
    forall h cs,
      extcall_spec
        "read_str" [] h
        (val.str cs) h

def output [io.interface] (s : string) : io unit :=
  do f ← io.mk_file_handle "OUT.imp" io.mode.append,
     io.fs.write f s.to_char_buffer,
     io.fs.write f "\n".to_char_buffer,
     io.fs.close f,
     io.print_ln s


def array_range : nat → list nat
| 0 := []
| (n + 1) := (n + 1) :: array_range n

def array_pretty' : nat → heap → int → option string
| 0 h _ := none
| (fuel + 1) h a :=
  if a < 0 ∨ int.of_nat h.length <= a
  then "fail: array_pretty bogus pointer"
  else
  match imp.heap.read h a with
  | some (imp.val.int n) :=
    if n < 0 ∨ ((int.of_nat h.length) <= a + n)
    then "fail: array_pretty bogus array"
    else
      (let ns := (array_range n.to_nat).reverse in
      let ovs := ns.map (fun i, imp.heap.read h (a + i)) in
      do vs <- monad.mapm (fun o, match o with
                                | some v := match v with
                                | (imp.val.addr a) := array_pretty' fuel h a
                                | v := to_string v
                                end
                                | none := "impossible"
                                end) ovs,
      some $ "[" ++ string.join (list.intersperse ", " vs) ++ "]")
  | _ := "array_pretty pointer to non-array"
  end

def array_pretty (h : imp.heap) (a : int) : string :=
(array_pretty' h.length h a).get_or_else "INTERNAL ERROR PRINTING ARRAY"

def extcall_io [io.interface] : string → list val → heap → io (val × heap)
| "print_val" (v :: []) h :=
  do match v with
     | (val.str s) := output s
     | (val.addr a) := output (array_pretty h a)
     | v := output (to_string v)
     end,
     pure (val.int 0, h)
| "read_int" [] h :=
  do ln ← io.get_line,
     match from_string ln.to_list with
     | none := io.fail $ "extcall: read_int could not parse " ++ ln
     | some i := pure (val.int i, h)
     end
| "read_bool" [] h :=
  do ln ← io.get_line,
     match ln with
     | "True" := pure (val.bool tt, h)
     | "False" := pure (val.bool ff, h)
     | s := io.fail $ "extcall: read_bool could not parse: " ++ s
     end
| "read_str" [] h :=
  do ln ← io.get_line,
     pure $ (val.str ln, h)
| _ _ h := pure (val.bool tt, h)
-- | "flush" [] h := (val.bool tt, h)
-- | "sleep" (val.int i :: []) _ := true
-- | "read_bool" [] _ := true
-- | "read_int" [] _ := true
-- | "read_str" [] _ := true
-- | _ _ _ :=

-- let extcall name args h =
--   match implode name, args with
--   | "print_val", [v] ->
--       output h v;
--       (Vint Big.zero, h)
--   | "flush", [] ->
--       flush stdout;
--       (Vint Big.zero, h)
--   | "sleep", [Vint i] ->
--       let ms = (Big_int.float_of_big_int i) /. 1000.0 in
--       ignore (Unix.select [] [] [] ms);
--       (Vint Big.zero, h)

--   | f, _ -> begin
--       prerr_endline ("extcall: bogus call to " ^ f);
--       (Vint Big.zero, h)
--     end

def extcall (ioi : io.interface) : string → list val → heap → val × heap
| s vs h := match unsafe_io ioi (@extcall_io ioi s vs h) with
| sum.inr e := (val.int 0, h)
| sum.inl v := v
end

axiom extcall_ok :
   forall ioi f vs h v' h',
     extcall ioi f vs h = (v', h') →
     extcall_spec f vs h v' h'

end imp
