import .syntax
import ..print_util

namespace imp

def val.to_string : val → string
| (val.bool tt) := "True"
| (val.bool ff) := "False"
| (val.int i) := to_string i
| (val.str s) := repr s
| (val.addr i) := i.repr

instance val.has_to_string : has_to_string val :=
⟨ val.to_string ⟩

open unary_op
def unary_op.to_string : unary_op → string
| neg := "-"
| not := "not"

instance unary_op.has_to_string : has_to_string unary_op :=
⟨ unary_op.to_string ⟩

open binary_op
def binary_op.to_string : binary_op → string
| add := "+"
| sub := "-"
| mul := "*"
| div := "/"
| mod := "%"
| eq := "=="
| lt := "<"
| le := "<="
| conj := "and"
| disj := "or"

instance binary_op.has_to_string : has_to_string binary_op :=
⟨ binary_op.to_string ⟩

def expr.to_string : expr → string
| (expr.val v) := to_string v
| (expr.var x) := x
| (expr.unary_op op e1) := op.to_string ++ " " ++ e1.to_string
| (expr.binary_op op e1 e2) := e1.to_string ++ " " ++ op.to_string ++ " " ++ e2.to_string
| (expr.length e) := "len(" ++ e.to_string ++ ")"
| (expr.index e1 e2) := e1.to_string ++ "[" ++ e2.to_string ++ "]"

instance expr.has_to_string : has_to_string expr :=
⟨ expr.to_string ⟩

def stmt.to_string_aux : stmt → list string
| (stmt.nop) := ["nop;"]
| (stmt.set x e) := [x ++ " = " ++ to_string e ++ ";"]
| (stmt.alloc x e1 e2) := [x ++ " = " ++ "alloc(" ++ to_string e1 ++ ", " ++ to_string e2 ++ ");"]
| (stmt.write x e1 e2) := [x ++ "[" ++ to_string e1 ++ "]" ++ " = " ++ to_string e2 ++ ";"]
| (stmt.call x fn args) := [x ++ " = " ++ fn ++ parens (concat "," (list.map expr.to_string args)) ++ ";"]
| (stmt.ifelse e p1 p2) :=
  ["if " ++ parens (to_string e) ++ " {"] ++ indent (stmt.to_string_aux p1) ++ ["} else {"] ++ indent (stmt.to_string_aux p2) ++ ["}"]
| (stmt.while g p1) := ["while " ++ parens (to_string g) ++ " {"] ++ indent (stmt.to_string_aux p1) ++ ["}"]
| (stmt.seq s1 s2) := stmt.to_string_aux s1 ++ stmt.to_string_aux s2

def stmt.to_string (s : stmt) :=
concat "\n" (stmt.to_string_aux s)

instance stmt.has_to_string : has_to_string stmt :=
⟨ stmt.to_string ⟩

def func.to_string : func → string
| (func.mk name ps body ret) :=
  "def " ++ name ++ parens (string.join $ list.intersperse "," ps) ++ " " ++
   curlies (concat "\n" $ indent $ body.to_string_aux ++ ["return " ++ to_string ret ++ ";"])

instance func.has_to_string : has_to_string func :=
⟨ func.to_string ⟩

def prog.to_string : prog → string
| (prog.mk funcs main ret) :=
  concat "\n\n" $ list.map func.to_string funcs ++ [main.to_string]

instance prog.has_to_string : has_to_string prog :=
⟨ prog.to_string ⟩



end imp
