import system.io

namespace imp

/- In a source program, a value corresponds to a constant. -/
inductive val : Type
| bool : bool → val
| int  : int → val
| str  : string → val
| addr : int → val

/- Both SugarImp and Imp have the same unary
   and binary operators, defined below in types
   `unary_op` and `binary_op` respectively. -/
inductive unary_op : Type
| neg
| not

inductive binary_op : Type
| add
| sub
| mul
| div
| mod
| eq
| lt
| le
| conj
| disj

/-
  In particular, expressions are pure (never have side effects),
  heap writes are directly through a local variable
  (no direct "multidimensional" array writes), and functions
  only (and always) return at the end.  We also have
  an additional kind of statement, [Snop], which does
  nothing.
-/
inductive expr : Type
| val : val → expr
| var : string → expr
| unary_op : unary_op → expr → expr
| binary_op : binary_op → expr → expr → expr
| length : expr → expr
| index : expr → expr → expr

inductive stmt : Type
| nop : stmt
| set : string → expr → stmt
| alloc : string → expr → expr → stmt
| write : string → expr → expr → stmt
| call : string → string → list expr → stmt
| ifelse : expr → stmt → stmt → stmt
| while : expr → stmt → stmt
| seq : stmt → stmt → stmt

inductive func : Type
/- Functions that always end with return, like
<<
  def f(x1, ..., xN):
    s
    return e
>> -/
| mk : string → list string → stmt → expr → func


inductive prog : Type
| mk : list func → stmt → expr → prog

@[reducible] def store : Type :=
list (string × val)

@[reducible] def heap : Type :=
list val

def store.empty : store := []

end imp
