import parsing.core
import parsing.char

namespace lexer

open parser
open parser.char

@[reducible] def parser (a : Type) := parser.char.t a

def range (start : char) (endc : char) : parser char :=
  parser.sat (λ c, start <= c ∧ c <= endc)

def choose {t : Type} : list (parser t) → parser t
| [] := parser.fail "choose: none of the choices succeded"
| (p :: ps) := p <|> choose ps

def ident : parser string := do
  let first_char := choose [range 'a' 'z', range 'A' 'Z', parser.sat (= '_')],
  let identp := many_char $ choose [range 'a' 'z', range 'A' 'Z', range '0' '9', parser.sat (= '_')],
  c ← first_char,
  id_rest ← identp,
  return $ (string.str "" c) ++ id_rest

-- this should fail if its not a number
def int.from_string_aux : int → list char → int
| accum [] := accum
| accum (c :: cs) := int.from_string_aux (accum * 10 + (c.to_nat - 48)) cs

def int.from_string (s : string) : int :=
  int.from_string_aux 0 s.to_list

-- def nonzero_integer_lit
def integer_lit : parser int :=
  int.from_string <$>
    ((string.str "" <$> parser.sat (= '0')) <|>
    (parser.many_char1 $ choose [range '1' '9', range '0' '9']))

open parser

-- and read_string buf = parse
--   | '"'       { STRLIT (Buffer.contents buf) }
--   | '\\' '"'  { Buffer.add_char buf '"'    ; read_string buf lexbuf }
--   | '\\' '\\' { Buffer.add_char buf '\\'   ; read_string buf lexbuf }
--   | '\\' 'b'  { Buffer.add_char buf '\b'   ; read_string buf lexbuf }
--   | '\\' 'f'  { Buffer.add_char buf '\012' ; read_string buf lexbuf }
--   | '\\' 'n'  { Buffer.add_char buf '\n'   ; read_string buf lexbuf }
--   | '\\' 'r'  { Buffer.add_char buf '\r'   ; read_string buf lexbuf }
--   | '\\' 't'  { Buffer.add_char buf '\t'   ; read_string buf lexbuf }
--   | [^ '"' '\\']+
--     { Buffer.add_string buf (Lexing.lexeme lexbuf);
--       read_string buf lexbuf
--     }
--   | _ as c
--     { raise (SyntaxError (mkstr "Illegal string character: %c" c)) }
--   | eof
--     { raise (SyntaxError ("Unterminated string literal")) }

def string_char : parser char :=
parser.sat (λc, c ≠ '\"' ∧ c ≠ '\\' ∧ c.val > 0x1f)
 <|>  (do str "\\",
         (str "t" >> return '\t') <|>
         (str "n" >> return '\n') <|>
         (str "\\" >> return '\\') <|>
         (str "\"" >> return '\"'))

def string_literal : parser string :=
str "\"" *> (many_char string_char) <* str "\""

def comment : parser string :=
parser.sat (= '#') *> many_char (parser.sat (≠ '\n'))

def whitespace : parser unit :=
parser.many' $ parser.one_of' " \t\x0d\n".to_list *> return ()

def ensure_ws := parser.many1 $ parser.one_of' " \t\x0d\n".to_list *> return ()

-- let line =
--   '\r' | '\n' | "\r\n"
def linebreak : parser unit :=
parser.sat (='\n') *> return ()

@[derive decidable_eq]
inductive tok
| true
| false
| len
| not
| add
| sub
| mul
| div
| mod
| eq
| lt
| le
| conj
| disj
| int_lit : ℤ → tok
| str_lit : string → tok
| nop
| assign
| semi
| if_t
| else_t
| while
| def_t
| ret_t
| alloc
| comma
| lparen
| rparen
| lcurly
| rcurly
| lsquare
| rsquare
| eof
| ident : string → tok

section tok_to_string

open tok

def tok.to_string : tok → string
| true := "true"
| false := "false"
| len := "len"
| not := "not"
| add := "add"
| sub := "sub"
| mul := "mul"
| div := "div"
| mod := "mod"
| eq := "eq"
| lt := "lt"
| le := "le"
| conj := "conj"
| disj := "disj"
| (int_lit i) := "int_lit " ++ to_string i
| (tok.str_lit s) := "str_lit " ++ to_string s
| nop := "nop"
| assign := "assign"
| semi := "semi"
| if_t := "if_t"
| else_t := "else_t"
| while := "while"
| def_t := "def_t"
| ret_t := "ret_t"
| alloc := "alloc"
| comma := "comma"
| lparen := "lparen"
| rparen := "rparen"
| lcurly := "lcurly"
| rcurly := "rcurly"
| lsquare := "lsquare"
| rsquare := "rsquare"
| eof := "eof"
| (tok.ident id) := id

instance tok.has_to_string : has_to_string tok :=
⟨ tok.to_string ⟩

end tok_to_string

universe u

def not {α : Type} : parser α → parser unit :=
fun p, (p >> parser.failure) <|> return ()

-- might still be bug prone essentially saying that the follow set of a keyword can be any non
-- ident letter
def keyword (s : string) : parser unit :=
try (str s <* (ensure_ws *> pure () <|> not ident))

-- Option is used to signal that we can skip things like whitespace, etc.
def single_token : parser (option tok) :=
  (keyword "True" *> pure tok.true) <|>
  (keyword "False" *> pure tok.false) <|>
  (some <$> tok.int_lit <$> integer_lit) <|>
  (some <$> tok.str_lit <$> string_literal) <|>
  -- operators
  (keyword "len" *> pure tok.len) <|>
  (keyword "not" *> pure tok.not) <|>
  (str "+" *> pure tok.add) <|>
  (str "-" *> pure tok.sub) <|>
  (str "*" *> pure tok.mul) <|>
  (str "/" *> pure tok.div) <|>
  (str "%" *> pure tok.mod) <|>
  (str "==" *> pure tok.eq) <|>
  (str "<=" *> pure tok.le) <|>
  (str "<" *> pure tok.lt) <|>
  (keyword "and" *> pure tok.conj) <|>
  (keyword "or" *> pure tok.disj) <|>
  -- statements
  (str "nop" *> pure tok.nop) <|>
  (str "=" *> pure tok.assign) <|>
  (keyword "alloc" *> pure tok.alloc) <|>
  (str ";" *> pure tok.semi) <|>
  (keyword "if" *> pure tok.if_t) <|>
  (keyword "else" *> pure tok.else_t) <|>
  (keyword "while" *> pure tok.while) <|>
  (keyword "def" *> pure tok.def_t) <|>
  (keyword "return" *> pure tok.ret_t) <|>
  -- misc
  (str "," *> pure tok.comma) <|>
  (str "(" *> pure tok.lparen) <|>
  (str ")" *> pure tok.rparen) <|>
  (str "{" *> pure tok.lcurly) <|>
  (str "}" *> pure tok.rcurly) <|>
  (str "[" *> pure tok.lsquare) <|>
  (str "]" *> pure tok.rsquare) <|>
  --  variables
  (some <$> tok.ident <$> ident) <|>
  (comment *> pure none) <|>
  (whitespace *> pure none) <|>
  (linebreak *> pure none) -- unknown character
  -- error
--   | _ as c
--     { raise (SyntaxError (mkstr "Unexpected char: %c" c)) }

def tokenize' : parser (list tok) :=
do ts ← many single_token,
   pure (list.filter_map id ts ++ [tok.eof])

def tokenize (buf : char_buffer) : list tok :=
match (many single_token).run buf with
| (sum.inl err_msg) := trace err_msg []
| (sum.inr opt_ts) := list.filter_map id opt_ts ++ [tok.eof]
end

end lexer
