import system.io
import parsing.core
import .imp
import .parser
import .to_python
import .gen

def DEFAULT_FUEL := 10000

def interp [ioi : io.interface] (fuel : nat) (p : imp.prog) : option (imp.heap × imp.val) :=
imp.interp.program ioi fuel p

def usage : string :=
string.join $ list.intersperse "\n" $ [
    "Usage: Imp [options] program.imp                            ",
    "                                                            ",
    "Interpret, print, or test Imp programs.                     ",
    "                                                            ",
    "OPTIONS:                                                    ",
    "  --interp    interpret Imp program                         ",
    "  --fuel N    interpret up to N steps                       ",
    "  --pretty    pretty print Imp program                      ",
    "  --python    print translation to Python                   ",
    "  --lean      print Lean AST                                 ",
    "  --test      compare interp and Python output              ",
    "  --rand N    print random Imp program (N controls size)    ",
    "  --help      display this usage information                ",
    "                                                            "]

@[derive decidable_eq]
inductive flag
| interp
| fuel : nat → flag
| pretty
| python
| lean
| test
| rand : nat → flag
| help

instance flag.inhabitated : inhabited flag :=
⟨ flag.help ⟩

namespace flag

def to_string : flag → string
| interp := "interp"
| (fuel n) := "fuel " ++ _root_.to_string n
| pretty := "pretty"
| python := "python"
| lean := "lean"
| test := "test"
| (rand n) := "rand " ++ _root_.to_string n
| help := "help"

end flag

instance flag.has_to_string : has_to_string flag :=
⟨ flag.to_string ⟩

variable [io.interface]

def parse_file {t : Type} (file_name : string) (p : parser char t) : io t :=
do buf ← io.fs.read_file file_name,
   match p.run buf with
   | sum.inl msg := io.fail msg
   | sum.inr v := return v
   end

private def parse_flags_aux : list string → option flag × (list flag) × (list string)
| (f :: fs) :=
  if f = "--interp"
  then   let (f, fs', ss) := parse_flags_aux fs in
     match f with
     | none := (some flag.interp, fs', ss)
     | some f' := (flag.interp, f' :: fs', ss)
     end
  else if f = "--pretty"
  then
     let (f, fs', ss) := parse_flags_aux fs in
     match f with
     | none := (some flag.pretty, fs', ss)
     | some f' := (flag.pretty, f' :: fs', ss)
     end
  else if f = "--python"
  then let (f, fs', ss) := parse_flags_aux fs in
     match f with
     | none := (some flag.python, fs', ss)
     | some f' := (flag.python, f' :: fs', ss)
     end
  else if f = "--lean"
  then let (f, fs', ss) := parse_flags_aux fs in
     match f with
     | none := (some flag.lean, fs', ss)
     | some f' := (flag.lean, f' :: fs', ss)
     end
  else if f = "--test"
  then let (f, fs', ss) := parse_flags_aux fs in
     match f with
     | none := (some flag.test, fs', ss)
     | some f' := (flag.test, f' :: fs', ss)
     end
  else if f = "--help"
  then let (f, fs', ss) := parse_flags_aux fs in
     match f with
     | none := (some flag.help , fs', ss)
     | some f' := (flag.help, f' :: fs', ss)
     end
  else if f = "--fuel"
  then
     have list.sizeof (list.tail fs) < string.length f + (1 + list.sizeof fs),
     from begin
        cases fs; simp [list.sizeof];
        well_founded_tactics.default_dec_tac,
     end,
     let (f, fs', ss) := parse_flags_aux (list.tail fs) in
     let fuel := ((from_string (list.head fs).to_list).get_or_else DEFAULT_FUEL).to_nat in
     match f with
     | none := (some $ flag.fuel fuel, fs', ss)
     | some f' := (flag.fuel fuel, f' :: fs', ss) --fix me
     end
  else if f = "--rand"
  then
     have list.sizeof (list.tail fs) < string.length f + (1 + list.sizeof fs),
     from begin
        cases fs; simp [list.sizeof];
        well_founded_tactics.default_dec_tac,
     end,
     let (f, fs', ss) := parse_flags_aux (list.tail fs) in
     let n := ((from_string (list.head fs).to_list).get_or_else 2).to_nat in
     match f with
     | none := (some $ flag.rand n, fs', ss) -- fix me
     | some f' := (flag.rand n, f' :: fs', ss) -- fix me
     end
  else
    let (of, fs', ss) := parse_flags_aux fs
    in (of, fs', f :: ss)
| [] := (none, [], [])

def parse_flags (fs : list string) : option flag × list flag × list string :=
parse_flags_aux fs

def for_each_file : list string → (string → io.handle → io unit) → io unit
| [] each := return ()
| (f :: fs) each :=
  do h ← io.mk_file_handle f io.mode.read,
     each f h,
     for_each_file fs each

def run_or_fail {a : Type} (p : parser char a) (out : char_buffer) : io a :=
match p.run out with
| sum.inl err := io.fail err
| sum.inr v := return v
end

def has_help (fs : list flag) : bool :=
  fs.any (λ f, match f with | flag.help := tt | _ := ff end)

def get_fuel : list flag → nat
| [] := DEFAULT_FUEL
| ((flag.fuel n) :: fs) := n
| (_ :: fs) := get_fuel fs


meta def print_result_val : (imp.heap × imp.val) → string
| (h, v) :=
  match v with
  | (imp.val.addr a) := imp.array_pretty h a
  | (imp.val.str str) := to_string str
  | v := to_string v
  end

def execute_python (orig_file : string) (p : imp.prog)  : io (string × string × int) :=
  let tmpfile := (orig_file ++ ".test.py") in
  let proc : io.process.spawn_args := {
    cmd := "python",
    args := [tmpfile],
    stdin := io.process.stdio.piped,
    stdout := io.process.stdio.piped,
    stderr := io.process.stdio.piped
  } in do outfile ← io.mk_file_handle tmpfile io.mode.write, -- fix me
          io.fs.write outfile (to_py_program p).to_char_buffer,
          io.fs.close outfile,
          child ← io.proc.spawn proc,
          -- another hack
          io.iterate () (fun _,
            do ln ← io.get_line,
               if ln = ""
               then return $ none
               else do io.fs.write child.stdin ln.to_char_buffer,
                       io.fs.write child.stdin "\n".to_char_buffer,
                       pure $ some ()),
          io.fs.close child.stdin,
          out ← io.fs.read_to_end child.stdout,
          err ← io.fs.read_to_end child.stderr,
          exitc ← io.proc.wait child,
          pure (out.to_string, err.to_string, exitc)

def get_output : io string :=
  do contents ← io.catch (io.fs.read_file "OUT.imp") (fun e1, pure "".to_char_buffer),
     io.cmd { cmd := "rm", args := ["-f", "OUT.imp"] },
     return contents.to_string

meta def results_match : string → option (imp.heap × imp.val) → (string × string × int) → bool
| imp_out imp_res (out, err, r) :=
  if r = 0
  then match imp_res with
       | some imp_res := let imp_str := imp_out ++ print_result_val imp_res ++ "\n"
                         in trace ("Python Output: " ++ out) (trace ("Imp Output: " ++ imp_str) imp_str = out)
       | _ := bool.ff
       end
  else if r = 1
  then match imp_res with
       | none := true
       | _ := false
       end
  else bool.ff

meta def main : io unit :=
do args ← io.cmdline_args,
   let (mode, flags, files) := parse_flags args,
  --  io.print_ln mode,
  --  io.print_ln flags,
  --  io.print_ln files,
   if files = [] ∨ has_help flags
   then io.print usage
   else for_each_file files $ λ file_name h,
     do out ← io.fs.read_to_end h,
        match mode.get_or_else flag.interp with
        | flag.interp :=
          do p ← run_or_fail imp.parser.program out,
             match interp (get_fuel flags) p with
             | none := io.fail "interpreter failed with: `none`"
             | some res := io.print_ln (print_result_val res)
             end
        | flag.pretty :=
          do p ← run_or_fail imp.parser.program out,
             io.print_ln (to_py_program p)
        | flag.python :=
          do p ← run_or_fail imp.parser.program out,
             io.print_ln (to_py_program p)
        | flag.lean :=
          do p ← run_or_fail imp.parser.program out,
             io.print_ln (repr p)
        | flag.test :=
          do p ← run_or_fail imp.parser.program out,
             let lean_res := interp (get_fuel flags) p,
             py_res ← execute_python file_name p,
             -- This is kind of a hack, in the future do a better job with this.
             lean_out ← get_output,
             -- io.print_ln "# Imp:\n\n%s\\n# Python:\n\n%s\n",
             let does_match := results_match lean_out lean_res py_res,
             if does_match
             then io.print_ln "PASS"
             else io.fail "FAIL"
        | (flag.rand size) :=
           do str ← to_string <$> imp.rand_prog size,
              io.print_ln str
        | flag.help := io.print_ln usage
        | _ := io.fail "unknown mode"
        end
