import .lexer
import .imp.syntax
import parsing
import parsing.expr

namespace imp
namespace parser

open parser
open lexer.tok

@[reducible] def t := parser lexer.tok

def is_ident : lexer.tok → bool
| (lexer.tok.ident _) := true
| _ := false

def is_int : lexer.tok → bool
| (lexer.tok.int_lit _) := true
| _ := false

def is_str : lexer.tok → bool
| (lexer.tok.str_lit _) := true
| _ := false

def ident : t string :=
do (lexer.tok.ident i) ← sat (λ t, is_ident t = tt) | parser.fail "not ident",
   return i

def int_lit : t int :=
do tk ← sat (λ t, is_int t = tt),
   match tk with
   | (lexer.tok.int_lit i) := return i
   | _ := parser.fail "expected integer literal"
   end

def str_lit : t string :=
do (lexer.tok.str_lit s) ← sat (λ t, is_str t = tt) | parser.fail "expected string",
   return s

def bracket {elem α : Type} : parser elem unit → parser elem unit → parser elem α → parser elem α :=
fun o c p, do o, r ← p, c, pure r

def curlies {α : Type} : t α → t α :=
bracket (sat (= lcurly) *> pure ()) (sat (= rcurly) *> pure ())

def parens {α : Type} (p : t α) : t α :=
bracket (sat (= lparen) *> pure ()) (sat (= rparen) *> pure ()) p

def squares {α : Type} (p : t α) : t α :=
sat (= lsquare) *> p <* sat (= rsquare)

def params : t (list string) :=
parens (sep_by (sat (= comma) *> pure ()) ident)

def binary_op (tk : lexer.tok) (o : binary_op) (assoc : parser.expr.assoc) : expr.operator lexer.tok imp.expr :=
parser.expr.operator.inf (do sat (= tk), return (imp.expr.binary_op o)) assoc

def unary_op (tk : lexer.tok) (o : unary_op) : expr.operator lexer.tok imp.expr :=
parser.expr.operator.pre (do sat (= tk), return (imp.expr.unary_op o))

section expression

parameter Exp : t imp.expr

def atomic_expr : t expr :=
  sat (= lexer.tok.true) *> pure (expr.val (val.bool tt)) <|>
  sat (= lexer.tok.false) *> pure (expr.val (val.bool ff)) <|>
  parens Exp <|>
  expr.val <$> val.int <$> int_lit <|>
  expr.val <$> val.str <$> str_lit <|>
  try (do e ← expr.var <$> ident,
   i ← squares Exp,
   pure (expr.index e i)) <|>
  try (do sat (= lexer.tok.len), e ← parens Exp, return (expr.length e)) <|>
  ident >>= λ i, pure (expr.var i)

open unary_op
open binary_op

def ops : expr.operator_table lexer.tok imp.expr := [
   [unary_op lexer.tok.sub neg, unary_op lexer.tok.not not],
   [binary_op lexer.tok.div div expr.assoc.left, binary_op lexer.tok.mul mul expr.assoc.left, binary_op lexer.tok.mod mod expr.assoc.left],
   [binary_op lexer.tok.add add expr.assoc.left, binary_op lexer.tok.sub sub expr.assoc.left],
   [binary_op lexer.tok.le le expr.assoc.none, binary_op lexer.tok.lt lt expr.assoc.none],
   [binary_op lexer.tok.eq eq expr.assoc.right],
   [binary_op lexer.tok.disj disj expr.assoc.left, binary_op lexer.tok.conj conj expr.assoc.none]
]

def binary_ops := expr.mk_expr_parser ops atomic_expr

end expression

def expression : t expr :=
fix (λ exp, binary_ops exp)

open has_map

def assign_to {res l r : Type} (lhs : t l) (rhs : t r) (cont : l → r → t res) : t res :=
    do l ← lhs,
       sat (= assign),
       r ← rhs,
       sat (= semi),
       cont l r

def alloc : t (imp.expr × imp.expr) :=
  do sat (= alloc),
     parens (do
        e1 ← expression,
        sat (= lexer.tok.comma),
        e2 ← expression,
        pure (e1, e2))

def call : t (string × list expr) :=
decorate_error "call expression" $
  do fn ← ident,
     args ← parens $ sep_by (sat (= comma) *> pure ()) expression,
     pure $ (fn, args)

def index_of : t (string × expr) :=
  do i ← ident,
     e ← squares $ expression,
     return (i, e)

def if_st (st : t stmt) : t stmt :=
  sat (= if_t) >> (imp.stmt.ifelse <$> parens expression <*> curlies st <*>
  ((sat (= else_t) *> curlies st) <|> pure stmt.nop))

def statement (st : t stmt) : t stmt :=
    try (assign_to ident expression (λ i e, pure $ imp.stmt.set i e)) <|>
    try (assign_to ident call (λ i ⟨ fn, args ⟩, pure $ imp.stmt.call i fn args)) <|>
    try (assign_to ident alloc (λ i ⟨ fn, args ⟩, pure $ imp.stmt.alloc i fn args)) <|>
    try (assign_to index_of expression (λ ⟨a, i⟩ e, pure $ imp.stmt.write a i e)) <|>
    if_st st <|>
    sat (= while) >> (imp.stmt.while <$> parens expression <*> curlies st) <|>
    sat (= nop) >> sat (= semi) >> pure imp.stmt.nop

def block : t stmt :=
parser.fix (fun st,
do ss ← many $ statement st,
   pure $ ss.foldr stmt.seq imp.stmt.nop)

def fn : t func :=
do sat (= lexer.tok.def_t),
   i ← ident,
   ps ← params,
   sat (= lcurly),
   body ← block,
   sat (= ret_t),
   ret ← expression,
   sat (= semi),
   sat (= rcurly),
   return (func.mk i ps body ret)

def file : t prog :=
do fns ← many fn,
   s ← block,
   sat (= ret_t),
   e ← expression,
   sat (= semi),
   sat (= lexer.tok.eof),
   return $ prog.mk fns s e

def pcompose {a b c d : Type} [has_to_string c] (f : b → buffer c) (p1 : parser a b) (p2 : parser c d) : parser a d :=
do res1 ← p1,
   match p2.run (f res1) with
   | sum.inl msg := parser.fail msg
   | sum.inr v := pure v
   end

def parse_buffer {α : Type} (s : char_buffer) (p : t α) : sum string α :=
let tokens := lexer.tokenize s
in trace (to_string tokens) (p.run tokens.to_buffer)

def program : parser char prog :=
pcompose (fun (l : list lexer.tok), l.to_buffer) lexer.tokenize' file

def parse_string {α : Type} (s : string) (p : t α) : sum string α :=
 let tokens := lexer.tokenize s.to_list.to_buffer
 in trace (to_string tokens) (p.run tokens.to_buffer)

end parser
end imp
