import ..imp.syntax
import ..imp.common

namespace imp

inductive eval.unary_op : unary_op → val →  val →  Prop
| neg :
    ∀ i, eval.unary_op unary_op.neg (val.int i)
                (val.int (- i))
| not :
    ∀ b,
      eval.unary_op unary_op.not (val.bool b) (val.bool (bnot b)).

open binary_op

inductive eval.binary_op : binary_op → val → val → val →  Prop
| add_i :
    forall i1 i2,
      eval.binary_op add (val.int i1) (val.int i2)
                 (val.int (i1 + i2))
| add_s :
    forall s1 s2,
      eval.binary_op add (val.str s1) (val.str s2)
                 (val.str $ s1 ++ s2)
| sub :
    forall i1 i2,
      eval.binary_op sub (val.int i1) (val.int i2)
                 (val.int $ i1 - i2)
| mul :
    forall i1 i2,
      eval.binary_op mul (val.int i1) (val.int i2)
                 (val.int $ i1 * i2)
| div :
    forall i1 i2,
      ¬ (i2 = 0) ->
      eval.binary_op div (val.int i1) (val.int i2)
                 (val.int $ int.fdiv i1 i2)
| mod :
    forall i1 i2,
      ¬ (i2 = 0) ->
      eval.binary_op mod (val.int i1) (val.int i2)
                 (val.int $ int.fmod i1 i2)
| eq :
    forall v1 v2,
      eval.binary_op eq v1 v2
                 (val.bool (decidable.to_bool $ v1 = v2))
| lt :
    forall i1 i2,
      eval.binary_op lt (val.int i1) (val.int i2)
                 (val.bool (decidable.to_bool $ i1 < i2))
| le :
    forall i1 i2,
      eval.binary_op le (val.int i1) (val.int i2)
                 (val.bool (decidable.to_bool $ i1 <= i2))
| conj :
    forall b1 b2,
      eval.binary_op conj (val.bool b1) (val.bool b2)
                 (val.bool (b1 && b2))
| disj :
    forall b1 b2,
      eval.binary_op disj (val.bool b1) (val.bool b2)
                (val.bool (b1 || b2))

inductive eval.expr (s : store) (h : heap) : expr -> val -> Prop
| val :
    forall v, eval.expr (expr.val v) v
| var :
    forall x v,
      s.lookup x = some v ->
      eval.expr (expr.var x) v
| unary_op :
    forall op e v v',
      eval.expr e v ->
      eval.unary_op op v v' ->
      eval.expr (expr.unary_op op e) v'
| binary_op :
    forall op e1 e2 v1 v2 v',
      eval.expr e1 v1 ->
      eval.expr e2 v2 ->
      eval.binary_op op v1 v2 v' ->
      eval.expr (expr.binary_op op e1 e2) v'
| length_a :
    forall e a l,
      eval.expr e (val.addr a) ->
      heap.read h a = some (val.int l) ->
      eval.expr (expr.length e) (val.int l)
| length_s :
    forall e cs l,
      eval.expr e (val.str cs) ->
      cs.length = l ->
      eval.expr (expr.length e) (val.int l)
| index_a :
    forall e1 e2 a l i v,
      eval.expr e1 (val.addr a) ->
      eval.expr e2 (val.int i) ->
      heap.read h a = some (val.int l) ->
      0 <= i /\ i < l ->
      h.read (a + i + 1) = some v ->
      eval.expr (expr.index e1 e2) v
| index_s :
    forall e1 e2 cs i c,
      eval.expr e1 (val.str cs) ->
      eval.expr e2 (val.int i) ->
      0 <= i →
      string.index cs (i.to_nat) = some c →
      eval.expr (expr.index e1 e2) (val.str $ string.str "" c)

inductive eval.seq_expr (s : store) (h : heap) :
  list expr -> list val -> Prop
| nil :
    eval.seq_expr [] []
| cons :
    forall e es v vs,
      eval.expr s h e v ->
      eval.seq_expr es vs ->
      eval.seq_expr (e :: es) (v :: vs).

open stmt

inductive eval.stmt (e : env) :
  store → heap → stmt → store → heap → Prop
| nop :
    forall s h,
      eval.stmt s h nop s h
| set :
    forall s h x e v,
      eval.expr s h e v ->
      eval.stmt s h (stmt.set x e) (update x v s) h
| ifelse_t :
    forall s h e p1 p2 s' h',
      eval.expr s h e (val.bool tt) ->
      eval.stmt s h p1 s' h' ->
      eval.stmt s h (ifelse e p1 p2) s' h'
| ifelse_f :
    forall s h e p1 p2 s' h',
      eval.expr s h e (val.bool ff) ->
      eval.stmt s h p2 s' h' ->
      eval.stmt s h (ifelse e p1 p2) s' h'
| while_t :
    forall s1 h1 e p s2 h2 s3 h3,
      eval.expr s1 h1 e (val.bool tt) ->
      eval.stmt s1 h1 p s2 h2 ->
      eval.stmt s2 h2 (while e p) s3 h3 ->
      eval.stmt s1 h1 (while e p) s3 h3
| while_f :
    forall s h e p,
      eval.expr s h e (val.bool ff) ->
      eval.stmt s h (while e p) s h
| seq :
    forall s1 h1 p1 s2 h2 p2 s3 h3,
      eval.stmt s1 h1 p1 s2 h2 ->
      eval.stmt s2 h2 p2 s3 h3 ->
      eval.stmt s1 h1 (seq p1 p2) s3 h3
| alloc :
    forall s h x e1 e2 i v,
      eval.expr s h e1 (val.int i) ->
      eval.expr s h e2 v ->
      0 <= i ->
      eval.stmt
        s h (alloc x e1 e2)
        (update x (val.addr h.length) s) (heap.alloc h i v)
| write :
    forall s h x e1 e2 a l i v h',
      store.lookup x s = some (val.addr a) ->
      heap.read h a = some (val.int l) ->
      eval.expr s h e1 (val.int i) ->
      eval.expr s h e2 v ->
      0 <= i /\ i < l ->
      heap.write h (a + i + 1) v = some h' ->
      eval.stmt s h (write x e1 e2) s h'
| call_internal :
    forall s h f params body ret vs sf sf' h' x es v,
      locate e f = some (func.mk f params body ret) ->
      eval.seq_expr s h es vs ->
      updates params vs store.empty = some sf ->
      eval.stmt
        sf h body
        sf' h' ->
      eval.expr sf' h' ret v ->
      eval.stmt
        s h (stmt.call x f es)
        (update x v s) h'
| eval_call_external :
    forall x f es s h vs v' h',
      locate e f = none ->
      eval.seq_expr s h es vs ->
      extcall_spec f vs h v' h' ->
      eval.stmt
        s h (stmt.call x f es)
        (update x v' s) h'

inductive eval.program : prog → heap → val → Prop
| eval_prog :
    forall funcs main ret s' h' v,
      eval.stmt funcs
        store.empty heap.empty main
        s' h' →
      eval.expr s' h' ret v →
      eval.program (prog.mk funcs main ret) h' v

end imp
